@@ -40,6 +40,7 @@
 
 #include <errno.h>
 #include <fcntl.h>
+#include <stdlib.h> /* for atoi */
 #include <string.h> /* for strcmp */
 #include <unistd.h> /* for chdir */
 #include <sys/stat.h>
@@ -125,8 +126,6 @@ parse_sftp_uri (GFile *file,
 {
 	char *tmp, *save;
 	char *uri;
-	char *u, *h, *s, *p;
-	char *h_end;
 
 	uri = g_file_get_uri (file);
 	g_assert (uri != NULL);
@@ -182,6 +181,9 @@ get_remote_ssh_command (const char *uri,
 
 	char *host_name, *path, *user_name;
 	char *command, *user_host, *unescaped_path;
+	char *quoted_path;
+	char *remote_command;
+	char *quoted_remote_command;
 	char *port_str;
 	guint host_port;
 
@@ -193,6 +195,7 @@ get_remote_ssh_command (const char *uri,
 
 	/* FIXME to we have to consider the remote file encoding? */
 	unescaped_path = g_uri_unescape_string (path, NULL);
+	quoted_path = g_shell_quote (unescaped_path);
 
 	port_str = NULL;
 	if (host_port != 0) {
@@ -208,19 +211,27 @@ get_remote_ssh_command (const char *uri,
 	}
 
 	if (command_to_run != NULL) {
-		command = g_strdup_printf ("ssh %s%s -t \"cd \'%s\' && %s\"", user_host, port_str, unescaped_path, command_to_run);
+		remote_command = g_strdup_printf ("cd %s && exec %s", quoted_path, command_to_run);
 	} else {
-		command = g_strdup_printf ("ssh %s%s -t \"cd \'%s\' && $SHELL -l\"", user_host, port_str, unescaped_path);
+		remote_command = g_strdup_printf ("cd %s && exec $SHELL -l", quoted_path);
 	}
 
-	g_free (user_name);
-	g_free (host_name);
-	g_free (path);
+	quoted_remote_command = g_shell_quote (remote_command);
 
+	command = g_strdup_printf ("ssh %s -t %s", user_host, quoted_remote_command);
+
+	g_free (user_name);
 	g_free (user_host);
-	g_free (unescaped_path);
+	g_free (host_name);
 	g_free (port_str);
 
+	g_free (path);
+	g_free (unescaped_path);
+	g_free (quoted_path);
+
+	g_free (remote_command);
+	g_free (quoted_remote_command);
+
 	return command;
 }
 
@@ -242,7 +253,7 @@ get_terminal_command_for_file_info (NautilusFileInfo *file_info,
 				    const char *command_to_run,
 				    gboolean remote_terminal)
 {
-	char *uri, *path;
+	char *uri, *path, *quoted_path;
 	char *command;
 
 	uri = nautilus_file_info_get_activation_uri (file_info);
@@ -284,11 +295,15 @@ get_terminal_command_for_file_info (NautilusFileInfo *file_info,
 	}
 
 	if (command == NULL && path != NULL) {
+		quoted_path = g_shell_quote (path);
+
 		if (command_to_run != NULL) {
-			command = g_strdup_printf ("cd '%s' && %s", path, command_to_run);
+			command = g_strdup_printf ("cd %s && exec %s", quoted_path, command_to_run);
 		} else {
-			command = g_strdup_printf ("cd '%s' && $SHELL -l", path);
+			command = g_strdup_printf ("cd %s && exec $SHELL -l", quoted_path);
 		}
+
+		g_free (quoted_path);
 	}
 
 	g_free (path);
@@ -489,11 +504,19 @@ nautilus_open_terminal_get_background_items (NautilusMenuProvider *provider,
 	uri = nautilus_file_info_get_activation_uri (file_info);
 	terminal_file_info = get_terminal_file_info (uri);
 
-	item = open_terminal_menu_item_new (file_info, terminal_file_info, gtk_widget_get_screen (window),
-					    NULL, terminal_file_info == FILE_INFO_SFTP, FALSE);
-	items = g_list_append (items, item);
+	if (terminal_file_info == FILE_INFO_SFTP ||
+	    terminal_file_info == FILE_INFO_DESKTOP ||
+	    uri_has_local_path (uri)) {
+		/* local locations or SSH */
+		item = open_terminal_menu_item_new (file_info, terminal_file_info, gtk_widget_get_screen (window),
+						    NULL, terminal_file_info == FILE_INFO_SFTP, FALSE);
+		items = g_list_append (items, item);
+	}
 
-	if (terminal_file_info == FILE_INFO_SFTP && uri_has_local_path (uri)) {
+	if ((terminal_file_info == FILE_INFO_SFTP ||
+	     terminal_file_info == FILE_INFO_OTHER) &&
+	    uri_has_local_path (uri)) {
+		/* remote locations that offer local back-mapping */
 		item = open_terminal_menu_item_new (file_info, terminal_file_info, gtk_widget_get_screen (window),
 						    NULL, FALSE, FALSE);
 		items = g_list_append (items, item);
@@ -501,6 +524,8 @@ nautilus_open_terminal_get_background_items (NautilusMenuProvider *provider,
 
 	if (display_mc_item () &&
 	    g_find_program_in_path ("mc") &&
+	    (terminal_file_info == FILE_INFO_DESKTOP &&
+	     (desktop_is_home_dir () || desktop_opens_home_dir ())) ||
 	    uri_has_local_path (uri)) {
 		item = open_terminal_menu_item_new (file_info, terminal_file_info, gtk_widget_get_screen (window), "mc", FALSE, FALSE);
 		items = g_list_append (items, item);
@@ -541,9 +566,11 @@ nautilus_open_terminal_get_file_items (NautilusMenuProvider *provider,
 		case FILE_INFO_LOCAL:
 		case FILE_INFO_SFTP:
 		case FILE_INFO_OTHER:
-			item = open_terminal_menu_item_new (files->data, terminal_file_info, gtk_widget_get_screen (window),
-							    NULL, terminal_file_info == FILE_INFO_SFTP, TRUE);
-			items = g_list_append (items, item);
+			if (terminal_file_info == FILE_INFO_SFTP || uri_has_local_path (uri)) {
+				item = open_terminal_menu_item_new (files->data, terminal_file_info, gtk_widget_get_screen (window),
+								    NULL, terminal_file_info == FILE_INFO_SFTP, TRUE);
+				items = g_list_append (items, item);
+			}
 
 			if (terminal_file_info == FILE_INFO_SFTP &&
 			    uri_has_local_path (uri)) {
